@@ -20,17 +20,19 @@ IMAP Library - a IMAP email testing library.
 """
 
 from email import message_from_bytes
+from email.header import decode_header
 from imaplib import IMAP4, IMAP4_SSL
 from re import findall
-from codecs import decode
 from time import sleep, time
+
 try:
     from urllib.request import urlopen
 except ImportError:
     from urllib2 import urlopen
 from builtins import str as ustr
 from ImapLibrary2.version import get_version
-
+from ImapLibrary2.imap_proxy import IMAP4Proxy, IMAP4SSLProxy
+import os.path
 __version__ = get_version()
 
 
@@ -131,11 +133,18 @@ class ImapLibrary2(object):
         if self._is_walking_multipart(email_index):
             body = self.get_multipart_payload(decode=True)
         else:
-            encoded_body = self._imap.uid('fetch', email_index, '(BODY[TEXT])')[1][0][1]
-            try:
-                body = decode(encoded_body, 'quopri_codec').decode('UTF-8')
-            except:
-                body = decode(encoded_body, 'quopri_codec').decode('ISO-8859-1')
+            encodedMessage = self._imap.uid('fetch', email_index, '(BODY[])')[1][0][1]
+            msg = message_from_bytes(encodedMessage)
+            if not msg.is_multipart():
+                body = msg.get_payload(decode=True).decode()
+            else:
+                # decode the email subject
+                subject, encoding = decode_header(msg["Subject"])[0]
+                if isinstance(subject, bytes):
+                    # if it's a bytes, decode to str
+                    subject = subject.decode(encoding)
+                raise Exception("get_email_body called on multipart email '%s'. Please first use method walk_multipart_email." % (subject))
+
         return body
 
     def get_links_from_email(self, email_index):
@@ -277,22 +286,74 @@ class ImapLibrary2(object):
         - ``port``: The IMAP port number. (Default None)
         - ``user``: The username to be use to authenticate mailbox on given ``host``.
         - ``folder``: The email folder to read from. (Default INBOX)
+        - ``proxy_host``: Proxy host to connect via. (Default None)
+        - ``proxy_port``: Proxy port to connect via. (Default None)
+        - ``proxy_user``: Proxy username to connect via. (Default None)
+        - ``proxy_password``: Proxy password to connect via. (Default None)
+        - ``proxy_type``: Proxy type to connect via. Available values are: http, socks4, socks5 (Default http)
 
         Examples:
         | Open Mailbox | host=HOST | user=USER | password=SECRET |
         | Open Mailbox | host=HOST | user=USER | password=SECRET | is_secure=False |
         | Open Mailbox | host=HOST | user=USER | password=SECRET | port=8000 |
         | Open Mailbox | host=HOST | user=USER | password=SECRET | folder=Drafts
+        | Open Mailbox | host=HOST | user=USER | password=SECRET | folder=Drafts | proxy_host=ProxyHost | proxy_port=8080 | proxy_username=ProxyUsername | proxy_password=ProxyPassword | proxy_type=http
         """
         host = kwargs.pop('host', kwargs.pop('server', None))
         is_secure = kwargs.pop('is_secure', 'True') == 'True'
         port = int(kwargs.pop('port', self.PORT_SECURE if is_secure else self.PORT))
         folder = '"%s"' % str(kwargs.pop('folder', self.FOLDER))
-        self._imap = IMAP4_SSL(host, port) if is_secure else IMAP4(host, port)
+        proxy_host = kwargs.pop('proxy_host', None)
+        proxy_port = kwargs.pop('proxy_port', None)
+        proxy_user = kwargs.pop('proxy_user', None)
+        proxy_password = kwargs.pop('proxy_password', None)
+        proxy_type = kwargs.pop('proxy_type', 'http')
+        if proxy_host != None and proxy_port != None:
+            if is_secure:
+                self._imap = IMAP4SSLProxy(
+                                host=host,
+                                port=port,
+                                proxy_host=proxy_host,
+                                proxy_port=int(proxy_port),
+                                proxy_type=proxy_type)
+            else:
+                self._imap = IMAP4Proxy(
+                                host=host,
+                                port=port,
+                                proxy_host=proxy_host,
+                                proxy_port=int(proxy_port),
+                                proxy_type=proxy_type)
+        else:
+            self._imap = IMAP4_SSL(host, port) if is_secure else IMAP4(host, port)
         self._imap.login(kwargs.pop('user', None), kwargs.pop('password', None))
         self._imap.select(folder)
         self._init_multipart_walk()
 
+    def open_mailbox_oauth(self, **kwargs):
+        """Open IMAP email client session to oauth provider with given ``user`` and ``access_token``.
+        Arguments:
+        - ``host``: The IMAP host server. (Default None)
+        - ``debug_level``: An integer from 0 to 5 where 0 disables debug output and 5 enables full output with wire logging and parsing logs. (Default 0)
+        - ``folder``: The email folder to read from. (Default INBOX)
+        - ``user``: The username (email address) of the account to authenticate.
+        - ``access_token``: An OAuth2 access token. Must not be base64-encoded, since imaplib does its own base64-encoding.
+
+        Examples:
+        | Open Mailbox | host=HOST | debug_level=2 | user=email@gmail.com | access_token=SECRET |
+        | Open Mailbox | host=HOST | debug_level=0 | user=email@gmail.com | access_token=SECRET | folder=Drafts
+        """
+        host = kwargs.pop('host', kwargs.pop('server', None))
+        debug_level = int(kwargs.pop('debug_level', 0))
+        folder = '"%s"' % str(kwargs.pop('folder', self.FOLDER))
+        user = str(kwargs.pop('user', None))
+        access_token = str(kwargs.pop('access_token', None))
+        access_string = 'user=%s\1auth=Bearer %s\1\1' % (user, access_token)
+        self._imap = IMAP4_SSL(host)
+        self._imap.debug = debug_level
+        self._imap.authenticate('XOAUTH2', lambda x: access_string)
+        self._imap.select(folder)
+        self._init_multipart_walk()
+
     def wait_for_email(self, **kwargs):
         """Wait for email message to arrived base on any given filter criteria.
         Returns email index of the latest email message received.
@@ -367,10 +428,18 @@ class ImapLibrary2(object):
         status, data = self._imap.select(folder)
         if status != 'OK':
             raise Exception("imap.select error: %s, %s" % (status, data))
+        subject = kwargs.pop('subject', None)
+        if subject:
+            self._imap.literal = subject.encode("utf-8")
         typ, msgnums = self._imap.uid('search', None, *criteria)
         if typ != 'OK':
             raise Exception('imap.search error: %s, %s, criteria=%s' % (typ, msgnums, criteria))
-        return msgnums[0].split()
+        if msgnums[0] is not None:
+            if type(msgnums[0]) != bytes:
+                return msgnums[0]
+            return msgnums[0].split()
+        else:
+            return []
 
     @staticmethod
     def _criteria(**kwargs):
@@ -388,14 +457,14 @@ class ImapLibrary2(object):
             criteria += ['FROM', '"%s"' % sender]
         if cc:
             criteria += ['CC', '"%s"' % cc]
-        if subject:
-            criteria += ['SUBJECT', '"%s"' % subject]
         if text:
             criteria += ['TEXT', '"%s"' % text]
         if status:
             criteria += [status]
         if not criteria:
             criteria = ['UNSEEN']
+        if subject:
+            criteria += ['SUBJECT']
         return criteria
 
     def _init_multipart_walk(self):
@@ -418,3 +487,36 @@ class ImapLibrary2(object):
         """Saves all existing emails to internal variable."""
         typ, mails = self._imap.uid('search', None, 'ALL')
         self._mails = mails[0].split()
+
+    def get_attachments_from_email (self, **kwargs) :
+        """Save attachments of email message on given ``email_index`` (overwrite if already exists).
+
+        Arguments:
+        - ``email_index``: An email index to identity the email message.
+        - ``target_folder`` : local folder for saving attachments to (needs to exist),
+            defaults to current directory if None
+
+        Examples:
+        | Get Attachments Email | INDEX | C:\\Users\\User\\test
+        """
+        email_index = kwargs.pop('email_index', None)
+        target_folder = kwargs.pop('target_folder', None)
+        attachments = []
+        data = self._imap.uid('fetch', email_index, '(RFC822)')[1][0][1]
+        msg = message_from_bytes(data)
+
+        if target_folder is None:
+            target_folder = './'
+
+        for part in msg.walk():
+            content_maintype = part.get_content_maintype()
+            filename = None
+            if content_maintype != "multipart":
+                filename = part.get_filename()
+                if filename :
+                    filepath = os.path.join(target_folder, filename)
+                    fp = open(filepath, 'wb')
+                    fp.write(part.get_payload(decode=True))
+                    fp.close()
+                    attachments.append(str(filepath))
+        return attachments
